/*:
 * @plugindesc [v1.6] Tự động xuống dòng theo từ + hỗ trợ mã điều khiển + đổi font toàn game (MV only)
 * @author ChatGPT
 *
 * @param Enable Auto Wrap
 * @text Xuống dòng tự động
 * @type boolean
 * @default true
 * @desc Bật hoặc tắt việc tự động xuống dòng theo từ.

 * @param Enable Code Support
 * @text Xuống dòng hỗ trợ mã điều khiển
 * @type boolean
 * @default true
 * @desc Xử lý \N[1], \V[1], \C[1]... khi tính chiều dài văn bản.

 * @param Change Font Settings
 * @text Thay đổi Font
 * @type boolean
 * @default false
 * @desc Bật nếu bạn muốn thiết lập font mặc định toàn game.

 * @param Font Settings
 * @text Cài đặt Font
 * @type struct<FontConfig>
 * @default {"Font Name":"GameFont","Font Size":"28"}
 * @parent Change Font Settings
 * @desc Thiết lập font nếu "Thay đổi Font" được bật.

 * @help
 * 📌 Plugin hỗ trợ:
 * - Tự động xuống dòng theo từ, không bị cắt giữa từ.
 * - Hỗ trợ mã điều khiển như \N[1], \V[2], \C[3], \I[5]...
 * - Có thể thay đổi font chữ toàn game nếu bật.
 *
 * 🔧 Dựa trên phương pháp của plugin OpheliaEnigma_ChangeFont.
 *
 * ⚠ Không hoàn toàn tương thích với plugin chỉnh sửa hệ thống hội thoại (như YEP_MessageCore).
 */

/*~struct~FontConfig:
 * @param Font Name
 * @text Tên font mặc định
 * @type string
 * @default GameFont
 * @desc Font chữ mặc định cho toàn bộ game. (VD: Arial, Tahoma...)

 * @param Font Size
 * @text Kích thước font mặc định
 * @type number
 * @default 28
 * @desc Cỡ chữ mặc định toàn game (đơn vị px).
 */

(function() {
  const parameters = PluginManager.parameters('AutoWordWrapByWord');
  const autoWrapEnabled = parameters['Enable Auto Wrap'] === 'true';
  const enableCodeSupport = parameters['Enable Code Support'] === 'true';
  const changeFont = parameters['Change Font Settings'] === 'true';

  let fontName = 'GameFont';
  let fontSize = 28;

  if (changeFont) {
    try {
      const fontSettings = JSON.parse(parameters['Font Settings']);
      fontName = fontSettings['Font Name'] || 'GameFont';
      fontSize = Number(fontSettings['Font Size'] || 28);
    } catch (e) {
      console.warn("AutoWordWrapByWord: Lỗi đọc cài đặt font.");
    }

    Window_Base.prototype.standardFontFace = function() {
      return fontName;
    };

    Window_Base.prototype.standardFontSize = function() {
      return fontSize;
    };
  }

  // Tự động xử lý văn bản xuống dòng
  const _Window_Message_processAllText = Window_Message.prototype.processAllText;

  Window_Message.prototype.processAllText = function(textState) {
    if (autoWrapEnabled) {
      let text = textState.text;

      if (enableCodeSupport) {
        text = this.convertEscapeCharactersForMeasure(text);
      }

      textState.text = this._autoWordWrap(text);
    }

    _Window_Message_processAllText.call(this, textState);
  };

  // Mã xử lý escape code cho đo chiều rộng
  Window_Base.prototype.convertEscapeCharactersForMeasure = function(text) {
    text = text.replace(/\\/g, '\x1b');
    text = text.replace(/\x1bV\[(\d+)\]/gi, (_, n) => $gameVariables.value(Number(n)));
    text = text.replace(/\x1bN\[(\d+)\]/gi, (_, n) => {
      const actor = $gameActors.actor(Number(n));
      return actor ? actor.name() : '';
    });
    text = text.replace(/\x1bP\[(\d+)\]/gi, (_, n) => {
      const member = $gameParty.members()[Number(n)];
      return member ? member.name() : '';
    });
    text = text.replace(/\x1bG/gi, TextManager.currencyUnit);
    text = text.replace(/\x1bC\[(\d+)\]/gi, '');
    text = text.replace(/\x1bI\[(\d+)\]/gi, '');
    text = text.replace(/\x1b{/g, '');
    text = text.replace(/\x1b}/g, '');
    return text;
  };

  // Tự động xuống dòng theo từ
  Window_Message.prototype._autoWordWrap = function(text) {
    const dummyBitmap = new Bitmap(1, 1);
    dummyBitmap.fontFace = this.standardFontFace();
    dummyBitmap.fontSize = this.standardFontSize();
    const maxWidth = this.contentsWidth();

    const resultLines = [];
    let currentLine = '';
    let currentWidth = 0;

    const words = text.match(/(\S+\s*)/g) || [];

    for (let word of words) {
      const wordWidth = dummyBitmap.measureTextWidth(word);
      if (currentWidth + wordWidth > maxWidth) {
        if (currentLine.length > 0) {
          resultLines.push(currentLine);
          currentLine = word;
          currentWidth = wordWidth;
        } else {
          resultLines.push(word);
          currentLine = '';
          currentWidth = 0;
        }
      } else {
        currentLine += word;
        currentWidth += wordWidth;
      }
    }

    if (currentLine.length > 0) {
      resultLines.push(currentLine);
    }

    return resultLines.join('\n');
  };
})();
